# -*- coding: utf-8 -*-
""" Wrapper for sme_synth.so C library """
import logging
import os
from os.path import normpath

import numpy as np

from .cwrapper import IDL_DLL, get_lib_name
from .smelib.libtools import get_full_datadir

logger = logging.getLogger(__name__)


class SME_DLL:
    """Object Oriented interface for the SME C library"""

    def __init__(self, libfile=None, datadir=None, state=None):
        #:LineList: Linelist passed to the library
        self.linelist = None
        #:int: Number of mu points passed to the library
        self.nmu = None
        #:Abund: Elemental abundances passed to the library
        self.abund = None
        #:float: First wavelength of the current segment in Angstrom
        self.wfirst = None
        #:float: Last wavelength of the current segment in Angstrom
        self.wlast = None
        #:float: Van der Waals broadening parameter set in the library
        self.vw_scale = None
        #:bool: Whether the library uses H2 broadening or not
        self.h2broad = False
        #:float: Effective temperature set in the model
        self.teff = None
        #:float: Surface gravity set in the model
        self.grav = None
        #:float: Turbulence velocity in the model in km/s
        self.vturb = None
        #:Atmo: Atmosphere structure in the model
        self.atmo = None
        #:dict: NLTE subgrids for nlte coefficient interpolation
        self.nlte_grids = {}
        self.ion = None

        self.lib = IDL_DLL(libfile)

        if self.interface == "Parallel":
            self.parallel = True
            if state is None:
                self.state = self.NewState()
            else:
                self.state = state
        else:
            self.parallel = False
            self.state = None

        self.SetLibraryPath(datadir)
        self.check_data_files_exist()

    def __del__(self):
        # Free the memory from the state when this is closed
        # self.FreeState()
        pass

    @property
    def libfile(self):
        """str: Location of the library file"""
        return self.lib.libfile

    @property
    def interface(self):
        return self.lib.interface

    @property
    def datadir(self):
        """str: Expected directory of the data files"""
        return self.GetLibraryPath()

    @property
    def ndepth(self):
        """int: Number of depth layers in the atmosphere model"""
        assert self.atmo is not None, "No model atmosphere has been set"
        motype = self.atmo.depth
        return len(self.atmo[motype])

    @property
    def nlines(self):
        """int: number of lines in the linelist"""
        assert self.linelist is not None, "No line list has been set"
        return len(self.linelist)

    @property
    def file(self):
        """str: (Expected) Location of the library file"""
        return get_lib_name()

    def check_data_files_exist(self):
        """
        Checks if required data files for the SME library exist.
        If they dont exist, SME will just segfault, without any hint.

        Raises
        ------
        FileNotFoundError
            If any of the files don't exist
        """
        names = self.GetDataFiles()
        directory = self.GetLibraryPath()
        for name in names:
            n = os.path.join(directory, name)
            if not os.path.exists(n):
                raise FileNotFoundError(
                    "Could not find required data file {name} in library directory {directory}".format(
                        name=name, directory=directory
                    )
                )

    def NewState(self, delete_old=True):
        # if delete_old and hasattr(self, "state") and self.state is not None:
        #     self.FreeState()
        self.state = self.lib.new_state()
        return self.state

    def FreeState(self):
        pass
        # if self.state is not None:
        #     self.lib.free_state(self.state)
        #     self.state = None

    def CopyState(self):
        if self.state is None:
            return None
        state = self.lib.copy_state(self.state)
        return state

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        cls = self.__class__
        state = self.CopyState()
        obj = cls(libfile=self.lib.libfile, datadir=self.datadir, state=state)
        obj.parallel = self.parallel
        obj.linelist = self.linelist
        obj.nmu = self.nmu
        obj.abund = self.abund
        obj.wfirst = self.wfirst
        obj.wlast = self.wlast
        obj.vw_scale = self.vw_scale
        obj.h2broad = self.h2broad
        obj.teff = self.teff
        obj.grav = self.grav
        obj.vturb = self.vturb
        obj.atmo = self.atmo
        obj.nlte_grids = self.nlte_grids
        obj.ion = self.ion
        return obj

    def SMELibraryVersion(self):
        """
        Return SME library version

        Returns
        -------
        version : str
            SME library version
        """
        version = self.lib.call(
            "SMELibraryVersion", raise_error=False, state=self.state
        )
        return version

    def GetLibraryPath(self):
        """Get the data file directory"""
        return self.lib.GetLibraryPath(raise_error=False, state=self.state)

    def GetDataFiles(self):
        """Get the required data files"""
        files = self.lib.GetDataFiles(raise_error=False, state=self.state)
        return files.split(";")

    def SetLibraryPath(self, libpath=None):
        """Set the path to the library"""
        if libpath is None:
            libpath = get_full_datadir()
        libpath = normpath(libpath) + os.sep
        self.lib.SetLibraryPath(libpath, type="string", state=self.state)

    def InputWaveRange(self, wfirst, wlast):
        """
        Read in Wavelength range

        Will raise an exception if wfirst is larger than wlast

        Parameters
        ----------
        wfirst : float
            first wavelength of the segment
        wlast : float
            last wavelength of the segment
        """
        assert (
            wfirst < wlast
        ), "Input Wavelength range is wrong, first wavelength is larger than last"
        self.lib.InputWaveRange(wfirst, wlast, type="double", state=self.state)

        self.wfirst = wfirst
        self.wlast = wlast

    def SetVWscale(self, gamma6):
        """
        Set van der Waals scaling factor

        Parameters
        ----------
        gamma6 : float
            van der Waals scaling factor
        """
        self.lib.SetVWscale(gamma6, type="double", state=self.state)
        self.vw_scale = gamma6

    def SetH2broad(self, h2_flag=True):
        """Set flag for H2 molecule"""
        if h2_flag:
            self.lib.SetH2broad(state=self.state)
            self.h2broad = True
        else:
            self.ClearH2broad()

    def ClearH2broad(self):
        """Clear flag for H2 molecule"""
        self.lib.ClearH2broad(state=self.state)
        self.h2broad = False

    def InputLineList(self, linelist):
        """
        Read in line list

        Parameters
        ---------
        atomic : array of size (nlines, 8)
            atomic linelist data for each line
            fields are: atom_number, ionization, wlcent, excit, gflog, gamrad, gamqst, gamvw
        species : array(string) of size (nlines,)
            names of the elements (with Ionization level)
        """
        try:
            atomic = linelist["atomic"].T
            species = linelist["species"]
        except AttributeError:
            raise TypeError("linelist has to be a LineList type")

        nlines = len(species)
        species = np.asarray(species, "U8")

        assert (
            atomic.shape[1] == nlines
        ), "Got wrong Linelist shape, expected ({nlines}, 8) but got {atomic}".format(
            nlines=nlines, atomic=atomic.shape
        )
        assert (
            atomic.shape[0] == 8
        ), "Got wrong Linelist shape, expected ({nlines}, 8) but got {atomic}".format(
            nlines=nlines, atomic=atomic.shape
        )

        if self.state is not None:
            self.state.contents.free_linelist()
            self.state.contents.free_opacities()

        self.lib.InputLineList(
            nlines,
            species,
            atomic,
            type=("int", "string", "double"),
            state=self.state,
        )

        self.linelist = linelist

    def OutputLineList(self):
        """
        Return line list

        Returns
        -------
        atomic : array of size (nlines, 6)
            relevant data of the linelist
            wlcent, excit, gflog, gamrad, gamqst, gamvw
        """
        nlines = self.nlines
        atomic = np.zeros((nlines, 6))
        self.lib.OutputLineList(
            nlines, atomic, type=("int", "double"), state=self.state
        )
        return atomic

    def UpdateLineList(self, atomic, species, index):
        """
        Change line list parameters

        Parameters
        ---------
        atomic : array of size (nlines, 8)
            atomic linelist data for each line
            fields are: atom_number, ionization, wlcent, excit, gflog, gamrad, gamqst, gamvw
        species : array(string) of size (nlines,)
            names of the elements (with Ionization level)
        index : array(int) of size (nlines,)
            indices of the lines to update relative to the overall linelist
        """
        nlines = atomic.shape[0]
        assert (
            atomic.shape[1] == 8
        ), "Got wrong Linelist shape, expected ({nlines}, 8) but got {atomic}".format(
            nlines=nlines, atomic=atomic.shape
        )

        assert (
            len(index) == nlines
        ), "Inconsistent number if lines, between index and linelist"
        assert (
            len(species) == nlines
        ), "Inconsistent number if lines, between index and linelist"

        atomic = atomic.T

        self.lib.UpdateLineList(
            nlines,
            species,
            atomic,
            index,
            type=("int", "str", "double", "short"),
            state=self.state,
        )

    def InputModel(self, teff, grav, vturb, atmo):
        """Read in model atmosphere

        Parameters
        ---------
        teff : float
            effective Temperature in Kelvin
        grav : float
            surface gravity in log10(cgs)
        vturb : float
            turbulence velocity in km/s
        atmo : Atmo
            atmosphere structure (see Atmo for details)
        """

        if teff <= 0:
            raise ValueError("Temperature must be positive (unit is Kelvin)")
        if vturb < 0:
            raise ValueError("Turbulence velocity must be positive or zero")

        try:
            motype = atmo["depth"]
            depth = atmo[motype]
            ndepth = len(depth)
            t = atmo["temp"]
            xne = atmo["xne"]
            xna = atmo["xna"]
            rho = atmo["rho"]
            vt = np.full(ndepth, vturb) if np.size(vturb) == 1 else vturb
            wlstd = atmo["wlstd"]
            opflag = atmo["opflag"]
            args = [
                ndepth,
                teff,
                grav,
                wlstd,
                motype,
                opflag,
                depth,
                t,
                xne,
                xna,
                rho,
                vt,
            ]
            type = "sdddusdddddd"  # s : short, d: double, u: unicode (string)

            if atmo["geom"] == "SPH":
                radius = atmo["radius"]
                height = atmo["height"]
                motype = "SPH"
                args = args[:4] + [motype, radius] + args[5:] + [height]
                type = type[:5] + "d" + type[5:] + "d"
        except AttributeError as ae:
            raise TypeError("atmo has to be an Atmo type, {ae}".format(ae=ae))

        if self.state is not None:
            self.state.contents.free_opacities()

        self.lib.InputModel(*args, type=type, state=self.state)

        self.teff = teff
        self.grav = grav
        self.vturb = vturb
        self.atmo = atmo

    def InputAbund(self, abund):
        """
        Pass abundances to radiative transfer code.

        Calculate elemental abundances from abundance pattern and metallicity.
        Metallicity adjustment is not applied to H or He.
        Renormalize abundances after applying metallicity.
        Introduced limiter in case the proposed step in abundance is too large.

        Parameters
        ---------
        abund : Abund
            abundance structure to be passed (see Abund for more details)
        """
        # Convert abundances to the right format
        # metallicity is included in the abundance class, ignored in function call
        abund = abund("sme", raw=True)
        assert isinstance(abund, np.ndarray)
        self.lib.InputAbund(abund, type="double", state=self.state)

        self.abund = abund

    def Opacity(self, getData=False, motype=1):
        """Calculate opacities

        Parameters
        ---------
        getData : bool
            if True copblu and copred (and copstd) will be returned
            requires that radiative transfer was run
        motype : int
            if getData is True and motype is 0 then copstd will also be returned

        Returns
        -------
        copblu : array of size (nmu,)
            only if getData is True
        copred : array of size (nmu,)
            only if getData is True
        copstd : array of size (nmu,)
            only if getData is True and motype is 0
        """
        args = []
        type = ""
        if getData:
            nmu = self.nmu
            copblu = np.zeros(nmu)
            copred = np.zeros(nmu)
            args = [nmu, copblu, copred]
            type = ["s", "d", "d"]

            if motype == 0:
                copstd = np.zeros(nmu)
                args += [copstd]
                type += ["d"]

        self.lib.Opacity(*args, type=type, state=self.state)

        return args[1:]

    def GetOpacity(self, switch, species=None, key=None):
        """
        Returns specific cont. opacity, different output depending on the input

        Parameters
        ----------
        switch : int

            :-3: COPSTD
            :-2: COPRED
            :-1: COPBLU
            :0: AHYD,
            :1: AH2P
            :2: AHMIN
            :3: SIGH
            :4: AHE1
            :5: AHE2,
            :6: AHEMIN
            :7: SIGHE,
            :8: ACOOL, continuous opacity C1, Mg1, Al1, Si1, Fe1, CH, NH, OH,
            :9: ALUKE, continuous opacity N1, O1, Mg2, Si2, Ca2,
            :10: AHOT
            :11: SIGEL
            :12: SIGH2

        """
        length = self.nmu
        result = np.ones(length)
        args = [switch, length, result]
        type = ["s", "s", "d"]

        if switch == 8:
            if species is not None:
                if key is None:
                    raise AttributeError(
                        "Both species and key keywords need to be set with switch 8, continous opacity"
                    )
                else:
                    args += [species, key]
                    type += ["u", "u"]
        elif switch == 9:
            if species is not None:
                args += [species]
                type += ["u"]

        self.lib.GetOpacity(*args, type=type, state=self.state)
        return result

    def Ionization(self, ion=0):
        """
        Calculate ionization balance for current atmosphere and abundances.
        Ionization state is stored in the external library.
        Ion is a bit flag with values (add them together to use multiple):

        :1: adopt particle number densities from EOS
        :2: adopt electron number densities from EOS
        :4: adopt gas densities (g/cm^3) from EOS

        instead of using values from model atmosphere. Different abundance patterns
        in the model atmosphere (usually scaled solar) and SME (may be non-solar)
        can affect line shape, e.g. shape of hydrogen lines.

        Parameters
        ----------
        ion : int
            flag that determines the behaviour of the C function
        """
        if self.state is not None:
            self.state.contents.free_ionization()

        self.lib.Ionization(
            ion,
            type="short",
            raise_error=False,
            raise_warning=True,
            state=self.state,
        )
        self.ion = ion

    def GetDensity(self):
        """
        Retrieve density in each layer

        Returns
        -------
        density : array of size (ndepth,)
            Density of the atmosphere in each layer
        """
        length = self.ndepth
        array = np.zeros(length, dtype=float)
        self.lib.GetDensity(length, array, type="sd", state=self.state)
        return array

    def GetNatom(self):
        """
        Get XNA

        Returns
        -------
        XNA : array of size (ndepth,)
            XNA in each layer
        """
        length = self.ndepth
        array = np.zeros(length, dtype=float)
        self.lib.GetNatom(length, array, type="sd", state=self.state)
        return array

    def GetNelec(self):
        """
        Get XNE (Electron number density) for each layer in the atmosphere

        Returns
        -------
        XNE : array of size (ndepth,)
            XNE in each layer
        """
        length = self.ndepth
        array = np.zeros(length, dtype=float)
        self.lib.GetNelec(length, array, type="sd", state=self.state)
        return array

    def Transf(
        self,
        mu,
        accrt,
        accwi,
        keep_lineop=False,
        long_continuum=True,
        nwmax=400000,
        wave=None,
    ):
        """
        Radiative Transfer Calculation

        Perform the radiative transfer calculation thorugh the atmosphere
        Requires that all parameters have been set beforehand

        Parameters
        ---------
        mu : array of shape (nmu,)
            mu angles (1 - cos(phi)) of different limb points along the stellar surface
        accrt : float
            accuracy of the radiative transfer integration
        accwi : float
            accuracy of the interpolation on the wavelength grid
        keep_lineop : bool, optional
            if True do not recompute the line opacities (default: False)
        long_continuum : bool, optional
            if True the continuum is calculated at every wavelength (default: True)
        nwmax : int, optional
            maximum number of wavelength points if wavelength grid is not set with wave (default: 400000)
        wave : array, optional
            wavelength grid to use for the calculation,
            if not set will use an adaptive wavelength grid with no constant step size (default: None)

        Returns
        -------
        nw : int
            number of actual wavelength points, i.e. size of wint_seg
        wint_seg : array of shape (nwave,)
            wavelength grid, the number of wavelengthpoints is equal to the number of lines * 2 - 1
            One point in the center of each line + plus one between the next line
        sint_seg : array of shape (nmu, nwave)
            spectrum for each mu point
        cint_seg : array of shape (nmu, nwave)
            continuum for each mu point
        """
        keep_lineop = 1 if keep_lineop else 0
        long_continuum = 1 if long_continuum else 0

        if wave is None:
            nw = 0
            wint_seg = np.zeros(nwmax)
        else:
            nwmax = nw = len(wave)
            wint_seg = np.asarray(wave)

        mu = np.asarray(mu)
        nmu = np.size(mu)

        # Prepare data:
        sint_seg = np.zeros((nwmax, nmu))  # line+continuum intensities
        cint_seg = np.zeros((nwmax, nmu))  # all continuum intensities
        cintr_seg = np.zeros(nmu)  # red continuum intensity

        type = "sdddiiddddss"  # s: short, d:double, i:int, u:unicode (string)

        self.lib.Transf(
            nmu,
            mu,
            cint_seg,
            cintr_seg,
            nwmax,
            nw,
            wint_seg,
            sint_seg,
            accrt,
            accwi,
            keep_lineop,
            long_continuum,
            type=type,
            state=self.state,
        )

        if nw == 0:
            nw = np.count_nonzero(wint_seg)

        from os.path import dirname, join

        from scipy.io import readsav

        examples_dir = join(dirname(__file__), "../../examples")
        sme2 = readsav(join(examples_dir, "sun_6440_grid_test.out"))

        wint_seg = wint_seg[:nw]
        sint_seg = sint_seg[:nw, :].T
        cint_seg = cint_seg[:nw, :].T

        sint_seg = np.nan_to_num(sint_seg, copy=False)
        cint_seg = np.nan_to_num(cint_seg, copy=False)

        self.nmu = nmu

        return nw, wint_seg, sint_seg, cint_seg

    def CentralDepth(self, mu, accrt):
        """
        This subroutine explicitly solves the transfer equation
        for a set of nodes on the star disk in the centers of spectral
        lines. The results are specific intensities.

        Parameters
        ----------
        mu : array of size (nmu,)
            mu values along the stellar disk to calculate
        accrt : float
            precision of the radiative transfer calculation

        Returns
        -------
        table : array of size (nlines,)
            Centeral depth (i.e. specific intensity) of each line
        """

        nmu = np.size(mu)
        nwsize = self.nlines
        table = np.zeros(nwsize)

        self.lib.CentralDepth(
            nmu, mu, nwsize, table, accrt, type="idifd", state=self.state
        )
        self.nmu = nmu

        return table

    def GetLineOpacity(self, wave):
        """
        Retrieve line opacity data from the C library

        Parameters
        ----------
        wave : float
            Wavelength of the line opacity to retrieve

        Returns
        ---------
        lop : array
            line opacity
        cop : array
            continuum opacity including scatter
        scr : array
            Scatter
        tsf : array
            Total source function
        csf : array
            Continuum source function
        """
        nmu = self.ndepth
        lop = np.zeros(nmu)
        cop = np.zeros(nmu)
        scr = np.zeros(nmu)
        tsf = np.zeros(nmu)
        csf = np.zeros(nmu)
        type = "dsddddd"
        self.lib.GetLineOpacity(
            wave, nmu, lop, cop, scr, tsf, csf, type=type, state=self.state
        )
        return lop, cop, scr, tsf, csf

    def GetLineRange(self):
        """Get the effective wavelength range for each line
        i.e. the wavelengths for which the line has significant impact

        Parameters
        ----------
        nlines : int
            number of lines in the linelist

        Returns
        -------
        linerange : array of size (nlines, 2)
            lower and upper wavelength for each spectral line
        """
        nlines = self.nlines
        linerange = np.zeros((nlines, 2))

        self.lib.GetLineRange(
            linerange, nlines, type=("double", "int"), state=self.state
        )

        return linerange

    def InputNLTE(self, bmat, lineindex):
        """
        Input NLTE departure coefficients

        Parameters
        ----------
        bmat : array of size (2, ndepth)
            departure coefficient matrix
        lineindex : float
            index of the line in the linelist
        """
        ndepth = self.ndepth
        nlines = self.nlines

        if not isinstance(bmat, (list, np.ndarray)):
            raise TypeError("Departure coefficient matrix is not an array")

        bmat = np.atleast_2d(bmat)
        if bmat.shape[0] != 2:
            raise ValueError(
                "Departure coefficient matrix has the wrong shape, expected (2, {ndepth}) but got {bmat} instead".format(
                    ndepth=ndepth, bmat=bmat.shape
                )
            )
        if bmat.shape[1] != ndepth:
            raise ValueError(
                "Departure coefficient matrix has the wrong shape, expected (2, {ndepth}) but got {bmat} instead".format(
                    ndepth=ndepth, bmat=bmat.shape
                )
            )

        if not isinstance(lineindex, (int, np.integer)):
            raise TypeError("Lineindex is not an integer type")

        if not 0 <= lineindex < nlines:
            raise ValueError(
                "Lineindex out of range, expected value between 0 and {nlines}, but got {lineindex} instead".format(
                    nlines=nlines, lineindex=lineindex
                )
            )

        self.lib.InputDepartureCoefficients(
            bmat, lineindex, type=("double", "int"), state=self.state
        )

    def GetNLTE(self, line):
        """Get the NLTE departure coefficients as stored in the C library

        Parameters
        ----------
        line : int
            requested line index, i.e. between 0 and number of lines

        Returns
        -------
        bmat : array of size (2, nrhox)
            departure coefficients for the given line index
        """
        nrhox = self.ndepth

        bmat = np.full((2, nrhox), -1.0, dtype=float)
        self.lib.GetDepartureCoefficients(
            bmat, nrhox, line, type=("double", "int", "int"), state=self.state
        )
        return bmat

    def ResetNLTE(self):
        """Reset departure coefficients from any previous call, to ensure LTE as default"""
        self.lib.ResetDepartureCoefficients(state=self.state)

    def GetNLTEflags(self):
        """Get an array that tells us which lines have been used with NLTE correction

        Parameters
        ----------
        linelist : int
            number of lines

        Returns
        -------
        nlte_flags : array(bool) of size (nlines,)
            True if line was used with NLTE, False if line is only LTE
        """
        nlines = self.nlines
        nlte_flags = np.zeros(nlines, dtype=np.int16)

        self.lib.GetNLTEflags(
            nlte_flags, nlines, type=("short", "int"), state=self.state
        )

        return nlte_flags.astype(bool)



# -*- coding: utf-8 -*-
# TODO implement atmosphere tests
import numpy as np
import pytest

from pysme.atmosphere.atmosphere import Atmosphere
from pysme.atmosphere.interpolation import AtmosphereInterpolator
from pysme.atmosphere.savfile import SavFile

from .test_largefilestorage import lfs_atmo, skipif_lfs


@pytest.fixture
def atmosphere_name():
    # TODO iterate over all possible options
    return "marcs2012p_t1.0.sav"


@pytest.fixture
def atmosphere(atmosphere_name):
    atmo = Atmosphere(source=atmosphere_name, method="grid", interp="TAU", depth="RHOX")
    return atmo


@pytest.fixture
@pytest.mark.usefixtures("lfs_atmo")
def interpolator(atmosphere, lfs_atmo):
    interp = AtmosphereInterpolator(
        depth=atmosphere.depth,
        interp=atmosphere.depth,
        geom=atmosphere.geom,
        lfs_atmo=lfs_atmo,
    )
    return interp


@pytest.fixture
@pytest.mark.usefixtures("lfs_atmo")
def atmosphere_grid(atmosphere_name, lfs_atmo):
    name = lfs_atmo.get(atmosphere_name)
    atmo = SavFile(name, source=name)
    return atmo


@skipif_lfs
@pytest.mark.usefixtures("lfs_atmo")
def test_grid_point(atmosphere_name, atmosphere_grid, lfs_atmo, interpolator):
    # TODO: get this values from the grid
    teff = 7000
    logg = 4
    monh = 0

    atmo_interp = interpolator.interp_atmo_grid(atmosphere_name, teff, logg, monh)
    atmo_grid = atmosphere_grid.get(teff, logg, monh)

    assert np.allclose(atmo_interp.temp, atmo_grid.temp[1:])
    assert np.allclose(atmo_interp.tau, atmo_grid.tau[1:])
    assert np.allclose(atmo_interp.rhox, atmo_grid.rhox[1:])
    assert np.allclose(atmo_interp.rho, atmo_grid.rho[1:])
    assert np.allclose(atmo_interp.xna, atmo_grid.xna[1:])
    assert np.allclose(atmo_interp.xne, atmo_grid.xne[1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import os
from os.path import dirname, exists, expanduser, join
from shutil import copy

from setuptools import setup

import versioneer

# Create folder structure for config files
print("Set up the configuration files for PySME")
directory = expanduser("~/.sme/")
conf = join(directory, "config.json")
atmo = join(directory, "atmospheres")
nlte = join(directory, "nlte_grids")
cache_atmo = join(atmo, "cache")
cache_nlte = join(nlte, "cache")

os.makedirs(directory, exist_ok=True)
os.makedirs(atmo, exist_ok=True)
os.makedirs(nlte, exist_ok=True)
os.makedirs(cache_atmo, exist_ok=True)
os.makedirs(cache_nlte, exist_ok=True)

# Create config file if it does not exist
if not exists(conf):
    # Hardcode default settings?
    defaults = {
        "data.file_server": "http://sme.astro.uu.se/atmos",
        "data.atmospheres": "~/.sme/atmospheres",
        "data.nlte_grids": "~/.sme/nlte_grids",
        "data.cache.atmospheres": "~/.sme/atmospheres/cache",
        "data.cache.nlte_grids": "~/.sme/nlte_grids/cache",
        "data.pointers.atmospheres": "datafiles_atmospheres.json",
        "data.pointers.nlte_grids": "datafiles_nlte.json",
    }

    # Save file to disk
    with open(conf, "w") as f:
        json.dump(defaults, f)
else:
    print("Configuration file already exists")

# Copy datafile pointers, for use in the GUI
print("Copy references to datafiles to config directory")
copy(
    join(dirname(__file__), "src/pysme/datafiles_atmospheres.json"),
    expanduser("~/.sme/datafiles_atmospheres.json"),
)
copy(
    join(dirname(__file__), "src/pysme/datafiles_nlte.json"),
    expanduser("~/.sme/datafiles_nlte.json"),
)

# TODO: Have smelib compiled before distribution
with open(join(dirname(__file__), "README.md"), "r") as fh:
    long_description = fh.read()

# Setup package
setup(
    name="pysme-astro",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Spectroscopy Made Easy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ansgar Wehrhahn, Jeff A. Valenti",
    author_email="ansgar.wehrhahn@physics.uu.se, valenti@stsci.edu",
    packages=[
        "pysme",
        "pysme.gui",
        "pysme.atmosphere",
        "pysme.linelist",
        "pysme.smelib",
    ],
    package_dir={"": "src"},
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=[
        "numpy",
        "scipy",
        "astropy",
        "pandas",
        "wget",
        "tqdm",
        "colorlog",
        "emcee",
        "pybtex",
        "flex-format",
    ],
    url="https://github.com/AWehrhahn/SME/",
    project_urls={
        "Bug Tracker": "https://github.com/AWehrhahn/SME/issues",
        "Documentation": "https://pysme-astro.readthedocs.io/en/latest/",
        "Source Code": "https://github.com/AWehrhahn/SME/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
)



# -*- coding: utf-8 -*-
import io
import logging
from collections.abc import Iterable
from numbers import Integral

import numpy as np
import numpy.lib.mixins
from flex.extensions.bindata import MultipleDataExtension
from flex.flex import FlexExtension

logger = logging.getLogger(__name__)

# TODO make this a proper subclass of np.ndarray
# see also https://numpy.org/devdocs/user/basics.subclassing.html
HANDLED_FUNCTIONS = {}


def implements(np_function):
    "Register an __array_function__ implementation for DiagonalArray objects."

    def decorator(func):
        HANDLED_FUNCTIONS[np_function] = func
        return func

    return decorator


class Iliffe_vector(numpy.lib.mixins.NDArrayOperatorsMixin, MultipleDataExtension):
    """
    Illiffe vectors are multidimensional (here 2D) but not necessarily rectangular
    Instead the index is a pointer to segments of a 1D array with varying sizes
    """

    def __init__(self, values, offsets=None, dtype=None):
        FlexExtension.__init__(self, cls=MultipleDataExtension)
        if offsets is not None:
            if offsets[0] != 0:
                offsets = [0, *offsets]
            self.data = np.asarray(values, dtype=dtype)
            self.offsets = np.asarray(offsets, dtype=int)
            return

        sizes = [len(v) for v in values]
        offsets = np.array([0, *np.cumsum(sizes)])

        data = np.concatenate(values)
        if dtype is not None:
            data = data.astype(dtype)

        self.data = data
        self.offsets = offsets

    def __getitem__(self, key):
        if isinstance(key, Integral):
            return self.__getsegment__(key)
        if isinstance(key, (slice, Iterable)) and not isinstance(key, tuple):
            if isinstance(key, slice):
                key = range(self.nseg)[key]
            values = [self.__getsegment__(k) for k in key]
            return self.__class__(values)
        if isinstance(key, tuple):
            if isinstance(key[0], Integral):
                return self[key[0]][key[1]]
            if isinstance(key[0], (slice, Iterable)):
                if isinstance(key[0], slice):
                    key0 = range(self.nseg)[key[0]]
                else:
                    key0 = key[0]
                values = [self.__getsegment__(k) for k in key0]
                values = [v[key[1]] for v in values]
                if isinstance(key[1], Integral):
                    return np.array(values)
                if isinstance(key[1], (slice, Iterable)):
                    return self.__class__(values)
        if isinstance(key, Iliffe_vector):
            data = [w[m] for w, m in zip(self.segments, key.segments)]
            return self.__class__(data)
        raise KeyError

    def __setitem__(self, key, value):
        isscalar = np.isscalar(value)
        if not isscalar:
            value = np.asarray(value)
        if isinstance(key, Integral):
            return self.__setsegment__(key, value)
        if isinstance(key, (slice, Iterable)) and not isinstance(key, tuple):
            if isinstance(key, slice):
                key = range(self.nseg)[key]
            if isscalar or (value.ndim == 1 and value.dtype != "O"):
                for k in key:
                    self.__setsegment__(k, value)
            else:
                for i, k in enumerate(key):
                    self.__setsegment__(k, value[i])
            return
        if isinstance(key, tuple):
            if isinstance(key[0], Integral):
                data = self.__getsegment__(key[0])
                data[key[1]] = value
                return
            if isinstance(key[0], (slice, Iterable)):
                if isinstance(key[0], slice):
                    key0 = range(self.nseg)[key[0]]
                else:
                    key0 = key[0]
                if isscalar or (value.ndim == 1 and value.dtype != "O"):
                    for k in key0:
                        data = self.__getsegment__(k)
                        data[key[1]] = value
                else:
                    for i, k in enumerate(key0):
                        data = self.__getsegment__(k)
                        data[key[1]] = value
                return
        if isinstance(key, Iliffe_vector):
            self.data[key.data] = value
            return
        raise KeyError

    def __getsegment__(self, seg):
        while seg < 0:
            seg = self.nseg - seg
        if seg > self.nseg - 1:
            raise IndexError
        low, upp = self.offsets[seg : seg + 2]
        return self.data[low:upp]

    def __setsegment__(self, seg, value):
        while seg < 0:
            seg = self.nseg - seg
        if seg > self.nseg - 1:
            raise IndexError
        low, upp = self.offsets[seg : seg + 2]
        if np.isscalar(value) or value.size == upp - low:
            # Keep using the existing memory
            self.data[low:upp] = value
        else:
            # Need to set new memory
            data = [s if i != seg else value for i, s in enumerate(self.segments)]
            self.offsets = np.array([0, *np.cumsum([len(s) for s in data])])
            self.data = np.concatenate(data)

    def __array__(self, dtype=None):
        arr = self.data
        if dtype is not None:
            arr = arr.astype(dtype)
        return arr

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if method == "__call__":
            inputs = list(inputs)
            for i, input in enumerate(inputs):
                if isinstance(input, self.__class__):
                    inputs[i] = input.__array__()
            if "out" in kwargs:
                out = []
                for i, o in enumerate(kwargs["out"]):
                    if isinstance(o, self.__class__):
                        out.append(o.__array__())
                    else:
                        out.append(o)
                kwargs["out"] = tuple(out)
            arr = ufunc(*inputs, **kwargs)
            return self.__class__.from_offsets(arr, self.offsets)
        else:
            return NotImplemented

    def __array_function__(self, func, types, args, kwargs):
        if func not in HANDLED_FUNCTIONS:
            return func(self.data, *args, **kwargs)
        # Note: this allows subclasses that don't override
        # __array_function__ to handle DiagonalArray objects.
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented
        return HANDLED_FUNCTIONS[func](*args, **kwargs)

    def __array_function_axis__(self, func, axis, args, kwargs):
        if axis is None:
            return func(self.data, *args, **kwargs)
        elif axis == 0:
            data = [func(s, *args, **kwargs) for s in self.segments]
            data = np.array(data)
            return data
        else:
            return NotImplemented

    def __repr__(self):
        return f"{self.__class__.__name__}({self.segments})"

    def __len__(self):
        return self.nseg

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def ndim(self):
        return 2

    @property
    def shape(self):
        return (self.nseg, self.sizes)

    @property
    def size(self):
        return self.data.size

    @property
    def sizes(self):
        return np.diff(self.offsets)

    @property
    def nseg(self):
        return len(self.offsets) - 1

    @property
    def segments(self):
        return [self.data[l:u] for l, u in zip(self.offsets[:-1], self.offsets[1:])]

    @implements(np.ravel)
    def ravel(self, *args, **kwargs):
        return np.ravel(self.data, *args, **kwargs)

    @implements(np.ndarray.flatten)
    def flatten(self, *args, **kwargs):
        return self.data.flatten(*args, **kwargs)

    @implements(np.copy)
    def copy(self, *args, **kwargs):
        data = np.copy(self.data, *args, **kwargs)
        offsets = np.copy(self.offsets, *args, **kwargs)
        return self.__class__(data, offsets=offsets)

    @implements(np.where)
    def where(self, *args, **kwargs):
        data = np.where(self.data, *args, **kwargs)
        res = self.__class__(data, offsets=self.offsets)
        return res

    @implements(np.all)
    def all(self, *args, axis=None, **kwargs):
        return self.__array_function_axis__(np.all, axis, args, kwargs)

    @implements(np.any)
    def any(self, *args, axis=None, **kwargs):
        return self.__array_function_axis__(np.any, axis, args, kwargs)

    @implements(np.min)
    def min(self, *args, axis=None, **kwargs):
        return self.__array_function_axis__(np.min, axis, args, kwargs)

    @implements(np.max)
    def max(self, *args, axis=None, **kwargs):
        return self.__array_function_axis__(np.max, axis, args, kwargs)

    @implements(np.mean)
    def mean(self, *args, axis=None, **kwargs):
        return self.__array_function_axis__(np.mean, axis, args, kwargs)

    @classmethod
    def from_offsets(cls, array, offsets):
        self = cls(array, offsets=offsets)
        return self

    @classmethod
    def from_indices(cls, array, indices):
        if indices is None:
            return cls([array])
        else:
            offsets = [0, *np.cumsum(indices)]
            arr = [array[l:u] for l, u in zip(offsets[:-1], offsets[1:])]
            return cls(arr)

    def astype(self, dtype):
        data = self.data.astype(dtype)
        offsets = self.offsets
        return Iliffe_vector(data, offsets, dtype=dtype)

    # For IO with Flex
    def _prepare(self, name: str):
        cls = self.__class__

        header_fname = f"{name}/header.json"
        header_info, header_bio = cls._prepare_json(header_fname, self.header)
        result = [(header_info, header_bio)]

        for key, value in enumerate(self.segments):
            data_fname = f"{name}/{key}.npy"
            data_info, data_bio = cls._prepare_npy(data_fname, value)
            result += [(data_info, data_bio)]

        return result

    @classmethod
    def _parse(cls, header: dict, members: dict):
        data = {key[:-4]: cls._parse_npy(bio) for key, bio in members.items()}
        data = [data[str(i)] for i in range(len(data))]
        ext = cls(values=data)
        return ext

    def to_dict(self):
        data = {"data": self.data, "offsets": self.offsets}
        return data

    @classmethod
    def from_dict(cls, data):
        obj = cls(values=data["data"], offsets=data["offsets"])
        return obj

    # def to_dict(self):
    #     cls = self.__class__
    #     obj = {"header": self.header}
    #     for i, v in enumerate(self.segments):
    #         obj[str(i)] = cls._np_to_dict(v)
    #     return obj

    # @classmethod
    # def from_dict(cls, header: dict, data: dict):
    #     data = {name: cls._np_from_dict(d) for name, d in data.items()}
    #     data = [data[str(i)] for i in range(len(data))]
    #     obj = cls(values=data)
    #     return obj

    def _save(self):
        data = {str(i): v for i, v in enumerate(self.segments)}
        ext = MultipleDataExtension(data=data)
        return ext

    @classmethod
    def _load(cls, ext: MultipleDataExtension):
        data = ext.data
        values = [data[str(i)] for i in range(len(data))]
        iv = cls(values=values)
        return iv

    def _save_v1(self, file, folder=""):
        """
        Creates a npz structure, representing the vector

        Returns
        -------
        data : bytes
            data to use
        """
        b = io.BytesIO()
        np.savez(b, *self.segments)
        file.writestr(f"{folder}.npz", b.getvalue())

    @classmethod
    def _load_v1(cls, file):
        # file: npzfile
        names = file.files
        values = [file[n] for n in names]
        return cls(values=values)



# -*- coding: utf-8 -*-
""" Wrapper for SME C library """

import logging
import os
from ctypes import cdll
from os.path import abspath, dirname, join, normpath
from posixpath import expanduser

import numpy as np

from .smelib.libtools import get_full_datadir, get_full_libfile

logger = logging.getLogger(__name__)

# Load the library
# This ensures that we can find the library when we load _smelib
libfile = get_full_libfile()
cdll.LoadLibrary(libfile)
from .smelib import _smelib

CURRENT_LIB = libfile


def reload_lib(libfile):
    global _smelib
    global CURRENT_LIB
    if libfile != CURRENT_LIB:
        del _smelib
        cdll.LoadLibrary(libfile)
        from .smelib import _smelib
    CURRENT_LIB = libfile


class SME_DLL:
    """Object Oriented interface for the SME C library"""

    def __init__(self, libfile=None, datadir=None):
        self.libfile = libfile
        reload_lib(libfile)

        self.SetLibraryPath(datadir)
        self.check_data_files_exist()

    @property
    def datadir(self):
        """str: Expected directory of the data files"""
        return self.GetLibraryPath()

    @datadir.setter
    def datadir(self, value):
        self.SetLibraryPath(value)

    @property
    def file(self):
        """str: (Expected) Location of the library file"""
        return libfile

    def check_data_files_exist(self):
        """
        Checks if required data files for the SME library exist.
        If they dont exist, SME will just segfault, without any hint.

        Raises
        ------
        FileNotFoundError
            If any of the files don't exist
        """
        names = self.GetDataFiles()
        directory = self.GetLibraryPath()
        for name in names:
            n = os.path.join(directory, name)
            if not os.path.exists(n):
                raise FileNotFoundError(
                    "Could not find required data file {name} in library directory {directory}".format(
                        name=name, directory=directory
                    )
                )

    def SMELibraryVersion(self):
        """Return SME library version"""
        return _smelib.LibraryVersion()

    def GetLibraryPath(self):
        """Get the data file directory"""
        return _smelib.GetLibraryPath()

    def GetDataFiles(self):
        """Get the required data files"""
        files = _smelib.GetDataFiles()
        return files.split(";")

    def SetLibraryPath(self, libpath=None):
        """Set the path to the library"""
        if libpath is None:
            libpath = get_full_datadir()
        libpath = normpath(libpath) + os.sep
        _smelib.SetLibraryPath(libpath)

    def InputWaveRange(self, wfirst, wlast):
        """
        Read in Wavelength range

        Will raise an exception if wfirst is larger than wlast

        Parameters
        ----------
        wfirst : float
            first wavelength of the segment
        wlast : float
            last wavelength of the segment
        """
        _smelib.InputWaveRange(wfirst, wlast)

    def SetVWscale(self, gamma6):
        """
        Set van der Waals scaling factor

        Parameters
        ----------
        gamma6 : float
            van der Waals scaling factor
        """
        _smelib.SetVWscale(gamma6)

    def SetH2broad(self, h2_flag=True):
        """Set flag for H2 molecule"""
        if h2_flag:
            _smelib.SetH2broad()
        else:
            _smelib.ClearH2broad()

    def ClearH2broad(self):
        """Clear flag for H2 molecule"""
        self.SetH2broad(False)

    def InputLineList(self, linelist):
        """
        Read in line list

        Parameters
        ---------
        atomic : array of size (nlines, 8)
            atomic linelist data for each line
            fields are: atom_number, ionization, wlcent, excit, gflog, gamrad, gamqst, gamvw
        species : array(string) of size (nlines,)
            names of the elements (with Ionization level)
        """

        # Check for unsupported lines in the linelist
        # The limits are given by which data is available in XSAHA in eos.f
        # The one letter elements need the space, as this is what we get from the
        # linelist species
        max_ion = {
            "H ": 2,
            "He": 3,
            "Li": 4,
            "Be": 4,
            "B ": 4,
            "C ": 6,
            "N ": 6,
            "O ": 6,
            "F ": 6,
            "Ne": 6,
            "Na": 6,
            "Mg": 6,
            "Al": 6,
            "Si": 6,
            "P ": 6,
            "S ": 6,
            "Cl": 5,
            "Ar": 5,
            "K ": 5,
            "Ca": 5,
            "Sc": 5,
            "Ti": 5,
            "V ": 5,
            "Cr": 5,
            "Mn": 5,
            "Fe": 5,
            "Co": 5,
            "Ni": 5,
            "Cu": 3,
            "Zn": 3,
            "Ga": 3,
            "Ge": 3,
            "As": 3,
            "Se": 3,
            "Br": 3,
            "Kr": 3,
            "Rb": 3,
            "Sr": 3,
            "Y ": 3,
            "Zr": 3,
            "Nb": 3,
            "Mo": 3,
            "Tc": 3,
            "Ru": 3,
            "Rh": 3,
            "Pd": 3,
            "Ag": 3,
            "Cd": 3,
            "In": 3,
            "Sn": 3,
            "Sb": 3,
            "Te": 3,
            "I ": 3,
            "Xe": 3,
            "Cs": 3,
            "Ba": 3,
            "La": 3,
            "Ce": 4,
            "Pr": 4,
            "Nd": 4,
            "Pm": 3,
            "Sm": 3,
            "Eu": 4,
            "Gd": 3,
            "Tb": 3,
            "Dy": 3,
            "Ho": 3,
            "Er": 3,
            "Tm": 3,
            "Yb": 3,
            "Lu": 3,
            "Hf": 3,
            "Ta": 3,
            "W ": 3,
            "Re": 3,
            "Os": 3,
            "Ir": 3,
            "Pt": 3,
            "Au": 3,
            "Hg": 3,
            "Tl": 3,
            "Pb": 3,
            "Bi": 3,
            "Po": 3,
            "At": 3,
            "Rn": 3,
            "Fr": 3,
            "Ra": 3,
            "Ac": 3,
            "Th": 3,
            "Pa": 3,
            "U ": 3,
            "Np": 3,
            "Pu": 3,
            "Am": 3,
            "Cm": 3,
            "Bk": 3,
            "Cf": 3,
            "Es": 3,
        }

        elem = linelist["species"].astype("U2")
        ion = linelist["ionization"]

        # For each element check if we are above the maximum ionization
        mask = np.full(len(linelist), False)
        for key, value in max_ion.items():
            mask |= (elem == key) & (ion > value)
        nlines = np.count_nonzero(mask)
        # And remove those lines
        if nlines > 0:
            warning = ", ".join(np.unique(linelist[mask]["species"]))
            logger.warning(
                "Discarding %i high ionization lines, of species %s",
                nlines,
                warning,
            )
            linelist = linelist[~mask]

        # Prepare arrays for the C code
        atomic = linelist["atomic"].T
        species = linelist["species"]
        species = np.asarray(species, "S8")

        _smelib.InputLineList(species, atomic)

    def OutputLineList(self):
        """
        Return line list

        Returns
        -------
        atomic : array of size (nlines, 6)
            relevant data of the linelist
            wlcent, excit, gflog, gamrad, gamqst, gamvw
        """
        return _smelib.OutputLineList()

    def UpdateLineList(self, atomic, species, index):
        """
        Change line list parameters

        Parameters
        ---------
        atomic : array of size (nlines, 8)
            atomic linelist data for each line
            fields are: atom_number, ionization, wlcent, excit, gflog, gamrad, gamqst, gamvw
        species : array(string) of size (nlines,)
            names of the elements (with Ionization level)
        index : array(int) of size (nlines,)
            indices of the lines to update relative to the overall linelist
        """
        index = np.asarray(index, dtype=np.int16)
        species = species[index].astype("S8")
        atomic = atomic[index].T

        _smelib.UpdateLineList(
            species,
            atomic,
            index,
        )

    def InputModel(self, teff, grav, vturb, atmo):
        """Read in model atmosphere

        Parameters
        ---------
        teff : float
            effective Temperature in Kelvin
        grav : float
            surface gravity in log10(cgs)
        vturb : float
            turbulence velocity in km/s
        atmo : Atmo
            atmosphere structure (see Atmo for details)
        """

        if teff <= 0:
            raise ValueError("Temperature must be positive (unit is Kelvin)")
        if vturb < 0:
            raise ValueError("Turbulence velocity must be positive or zero")

        motype = atmo["depth"]
        depth = atmo[motype]
        ndepth = len(depth)
        temp = atmo["temp"]
        xne = atmo["xne"]
        xna = atmo["xna"]
        rho = atmo["rho"]
        vt = np.full(ndepth, vturb) if np.size(vturb) == 1 else vturb
        wlstd = atmo["wlstd"]
        opflag = atmo["opflag"].astype(np.int16)

        kwargs = {}
        if atmo["geom"] == "SPH":
            kwargs["radius"] = atmo["radius"]
            kwargs["height"] = atmo["height"]
            motype = "SPH"

        _smelib.InputModel(
            teff, grav, wlstd, motype, opflag, depth, temp, xne, xna, rho, vt, **kwargs
        )

    def InputAbund(self, abund):
        """
        Pass abundances to radiative transfer code.

        Calculate elemental abundances from abundance pattern and metallicity.
        Metallicity adjustment is not applied to H or He.
        Renormalize abundances after applying metallicity.
        Introduced limiter in case the proposed step in abundance is too large.

        Parameters
        ---------
        abund : Abund
            abundance structure to be passed (see Abund for more details)
        """
        abund = abund("sme", raw=True)
        # TODO: does this make sense or should we fix the C code?
        abund[np.isnan(abund)] = -99
        abund[abund < -99] = -99

        _smelib.InputAbund(abund)

    def Opacity(self):
        """Calculate opacities"""
        _smelib.Opacity()

    def GetOpacity(self, switch, species=None, key=None):
        """
        Returns specific cont. opacity, different output depending on the input

        Parameters
        ----------
        switch : str
            one of [COPSTD, COPRED, COPBLU, AHYD, AH2P, AHMIN, SIGH, AHE1, AHE2,
            AHEMIN, SIGHE, ACOOL, ALUKE, AHOT, SIGEL, SIGH2]
        key : str, optional
            for ACOOL, one of [new, old, fraction]
        species : str, optional
            for ACOOL and ALUKE it specifies the element
            ACOOL: C1, Mg1, Al1, Si1, Fe1, CH, NH, OH
            ALUKE: N1, O1, Mg2, Si2, Ca2
        """
        kwargs = {}
        if key is not None:
            kwargs["key"] = key
        if species is not None:
            kwargs["species "] = species
        return _smelib.GetOpacity(switch, **kwargs)

    def Ionization(self, ion=0):
        """
        Calculate ionization balance for current atmosphere and abundances.
        Ionization state is stored in the external library.
        Ion is a bit flag with values (add them together to use multiple):

        :1: adopt particle number densities from EOS
        :2: adopt electron number densities from EOS
        :4: adopt gas densities (g/cm^3) from EOS

        instead of using values from model atmosphere. Different abundance patterns
        in the model atmosphere (usually scaled solar) and SME (may be non-solar)
        can affect line shape, e.g. shape of hydrogen lines.

        Parameters
        ----------
        ion : int
            flag that determines the behaviour of the C function
        """
        _smelib.Ionization(ion)

    def GetDensity(self):
        """
        Retrieve density in each layer

        Returns
        -------
        density : array of size (ndepth,)
            Density of the atmosphere in each layer
        """
        return _smelib.GetDensity()

    def GetNatom(self):
        """
        Get XNA

        Returns
        -------
        XNA : array of size (ndepth,)
            XNA in each layer
        """
        return _smelib.GetNatom()

    def GetNelec(self):
        """
        Get XNE (Electron number density) for each layer in the atmosphere

        Returns
        -------
        XNE : array of size (ndepth,)
            XNE in each layer
        """
        return _smelib.GetNelec()

    def Transf(
        self,
        mu,
        wave=None,
        nwmax=400000,
        accrt=1e-3,
        accwi=3e-3,
        keep_lineop=False,
        long_continuum=True,
    ):
        """
        Radiative Transfer Calculation

        Perform the radiative transfer calculation thorugh the atmosphere
        Requires that all parameters have been set beforehand

        Parameters
        ---------
        mu : array of shape (nmu,)
            mu angles (1 - cos(phi)) of different limb points along the stellar surface
        accrt : float
            accuracy of the radiative transfer integration
        accwi : float
            accuracy of the interpolation on the wavelength grid
        keep_lineop : bool, optional
            if True do not recompute the line opacities (default: False)
        long_continuum : bool, optional
            if True the continuum is calculated at every wavelength (default: True)
        nwmax : int, optional
            maximum number of wavelength points if wavelength grid is not set with wave (default: 400000)
        wave : array, optional
            wavelength grid to use for the calculation,
            if not set will use an adaptive wavelength grid with no constant step size (default: None)

        Returns
        -------
        nw : int
            number of actual wavelength points, i.e. size of wint_seg
        wint_seg : array of shape (nwave,)
            wavelength grid, the number of wavelengthpoints is equal to the number of lines * 2 - 1
            One point in the center of each line + plus one between the next line
        sint_seg : array of shape (nmu, nwave)
            spectrum for each mu point
        cint_seg : array of shape (nmu, nwave)
            continuum for each mu point
        """
        keep_lineop = 1 if keep_lineop else 0
        long_continuum = 1 if long_continuum else 0

        # keywords = {"mu", "wave", "nwmax", "accrt", "accwi", "keep_lineop", "long_continuum"}
        nw, wave, sint, cint = _smelib.Transf(
            mu, wave, nwmax, accrt, accwi, keep_lineop, long_continuum
        )

        # Resize the arrays
        wave = wave[:nw]
        sint = sint[:nw, :].T
        cint = cint[:nw, :].T

        sint = np.nan_to_num(sint, copy=False)
        cint = np.nan_to_num(cint, copy=False)

        return nw, wave, sint, cint

    def CentralDepth(self, mu, accrt):
        """
        This subroutine explicitly solves the transfer equation
        for a set of nodes on the star disk in the centers of spectral
        lines. The results are specific intensities.

        Parameters
        ----------
        mu : array of size (nmu,)
            mu values along the stellar disk to calculate
        accrt : float
            precision of the radiative transfer calculation

        Returns
        -------
        table : array of size (nlines,)
            Centeral depth (i.e. specific intensity) of each line
        """

        return _smelib.CentralDepth(mu, accrt)

    def GetLineOpacity(self, wave):
        """
        Retrieve line opacity data from the C library

        Parameters
        ----------
        wave : float
            Wavelength of the line opacity to retrieve

        Returns
        ---------
        lop : array
            line opacity
        cop : array
            continuum opacity including scatter
        scr : array
            Scatter
        tsf : array
            Total source function
        csf : array
            Continuum source function
        """
        lop, cop, scr, tsf, csf = _smelib.GetLineOpacity(wave)
        return lop, cop, scr, tsf, csf

    def GetLineRange(self):
        """Get the effective wavelength range for each line
        i.e. the wavelengths for which the line has significant impact

        Returns
        -------
        linerange : array of size (nlines, 2)
            lower and upper wavelength for each spectral line
        """
        return _smelib.GetLineRange()

    def InputDepartureCoefficients(self, bmat, lineindex):
        """
        Input NLTE departure coefficients

        Parameters
        ----------
        bmat : array of size (2, ndepth)
            departure coefficient matrix
        lineindex : float
            index of the line in the linelist
        """
        bmat = np.atleast_2d(bmat)
        _smelib.InputDepartureCoefficients(bmat, lineindex)

    def GetDepartureCoefficients(self, line):
        """Get the NLTE departure coefficients as stored in the C library

        Parameters
        ----------
        line : int
            requested line index, i.e. between 0 and number of lines

        Returns
        -------
        bmat : array of size (2, nrhox)
            departure coefficients for the given line index
        """
        return _smelib.GetDepartureCoefficients(line)

    def ResetDepartureCoefficients(self):
        """Reset departure coefficients from any previous call, to ensure LTE as default"""
        _smelib.ResetDepartureCoefficients()

    def GetNLTEflags(self):
        """Get an array that tells us which lines have been used with NLTE correction

        Returns
        -------
        nlte_flags : array(bool) of size (nlines,)
            True if line was used with NLTE, False if line is only LTE
        """
        nlte_flags = _smelib.GetNLTEflags()
        return nlte_flags.astype(bool)



# -*- coding: utf-8 -*-
"""
Handles abstract LineList data
Implementation of a specific source (e.g. Vald) should be in its own file

Uses a pandas dataframe under the hood to handle the data
"""
import io
import json
import logging

import numpy as np
import pandas as pd
from flex.extensions.tabledata import JSONTableExtension
from scipy import constants

from ..persistence import IPersist
from ..util import air2vac, vac2air

logger = logging.getLogger(__name__)


class LineListError(Exception):
    """Raise when attempt to read a line data file fails"""


class LineList(IPersist):
    """Atomic data for a list of spectral lines"""

    _base_columns = [
        "species",
        "wlcent",
        "excit",
        "gflog",
        "gamrad",
        "gamqst",
        "gamvw",
        "atom_number",
        "ionization",
    ]
    string_columns = ["species", "term_lower", "term_upper", "reference"]

    # Citations are added in the submodule (e.g. ValdFile)
    citation_info = ""

    @staticmethod
    def parse_line_error(error_flags, values=None):
        """Transform Line Error flags into relative error values

        Parameters
        ----------
        error_flags : list(str)
            Error Flags as defined by various references
        values : float
            depths of the lines, to convert absolute to relative errors

        Returns
        -------
        errors : list(float)
            relative errors for each line
        """
        if values is None:
            values = np.ones(len(error_flags))

        nist = {
            "AAA": 0.003,
            "AA": 0.01,
            "A+": 0.02,
            "A": 0.03,
            "B+": 0.07,
            "B": 0.1,
            "C+": 0.18,
            "C": 0.25,
            "C-": 0.3,
            "D+": 0.4,
            "D": 0.5,
            "D-": 0.6,
            "E": 0.7,
        }
        error = np.ones(len(error_flags), dtype=float)
        for i, (flag, _) in enumerate(zip(error_flags, values)):
            if len(flag) == 0:
                error[i] = 0.5
            elif flag[0] in [" ", "_", "P"]:
                # undefined or predicted
                error[i] = 0.5
            elif flag[0] == "E":
                # absolute error in dex
                # TODO absolute?
                error[i] = 10 ** float(flag[1:])
            elif flag[0] == "C":
                # Cancellation Factor, i.e. relative error
                try:
                    error[i] = abs(float(flag[1:]))
                except ValueError:
                    error[i] = 0.5
            elif flag[0] == "N":
                # NIST quality class
                flag = flag[1:5].strip()
                try:
                    error[i] = nist[flag]
                except KeyError:
                    error[i] = 0.5
        return error

    @staticmethod
    def guess_format(kwargs):
        short_line_format = kwargs.pop(
            "short_line_format", kwargs.pop("short_format", None)
        )
        if short_line_format is not None:
            return short_line_format

        keys = kwargs.keys()
        if (
            "line_extra" in keys
            and "line_lulande" in keys
            and "line_term_low" in keys
            and "line_term_upp" in keys
        ):
            return 2
        return 1

    @classmethod
    def from_IDL_SME(cls, **kwargs):
        """extract LineList from IDL SME structure keywords"""
        species = kwargs.pop("species").astype("U")
        atomic = np.asarray(kwargs.pop("atomic"), dtype="<f8")
        lande = np.asarray(kwargs.pop("lande"), dtype="<f8")
        depth = np.asarray(kwargs.pop("depth"), dtype="<f8")
        lineref = kwargs.pop("lineref").astype("U")
        short_line_format = cls.guess_format(kwargs)
        if short_line_format == 2:
            line_extra = np.asarray(kwargs.pop("line_extra"), dtype="<f8")
            line_lulande = np.asarray(kwargs.pop("line_lulande"), dtype="<f8")
            line_term_low = kwargs.pop("line_term_low").astype("U")
            line_term_upp = kwargs.pop("line_term_upp").astype("U")

        # If there is only one line, it is 1D in the IDL structure, but we expect 2D
        atomic = np.atleast_2d(atomic)

        data = {
            "species": species,
            "atom_number": atomic[:, 0],
            "ionization": atomic[:, 1],
            "wlcent": atomic[:, 2],
            "excit": atomic[:, 3],
            "gflog": atomic[:, 4],
            "gamrad": atomic[:, 5],
            "gamqst": atomic[:, 6],
            "gamvw": atomic[:, 7],
            "lande": lande,
            "depth": depth,
            "reference": lineref,
        }

        if short_line_format == 1:
            lineformat = "short"
        if short_line_format == 2:
            lineformat = "long"
            error = [s[0:11].strip() for s in lineref]
            error = LineList.parse_line_error(error, depth)
            data["error"] = error
            data["lande_lower"] = line_lulande[:, 0]
            data["lande_upper"] = line_lulande[:, 1]
            data["j_lo"] = line_extra[:, 0]
            data["e_upp"] = line_extra[:, 1]
            data["j_up"] = line_extra[:, 2]
            data["term_lower"] = [t[10:].strip() for t in line_term_low]
            data["term_upper"] = [t[10:].strip() for t in line_term_upp]

        linedata = pd.DataFrame.from_dict(data)

        return (linedata, lineformat)

    def __init__(self, linedata=None, lineformat="short", medium=None, **kwargs):
        if linedata is None or len(linedata) == 0:
            if isinstance(linedata, pd.DataFrame):
                linedata = pd.DataFrame(data=linedata, columns=self._base_columns)
            elif "atomic" in kwargs.keys():
                # everything is in the kwargs (usually by loading from old SME file)
                linedata, lineformat = LineList.from_IDL_SME(**kwargs)
            else:
                linedata = pd.DataFrame(data=[], columns=self._base_columns)
        else:
            if isinstance(linedata, LineList):
                linedata = linedata._lines
                lineformat = linedata.lineformat
                medium = linedata._medium
            else:
                if isinstance(linedata, (list, np.ndarray)):
                    # linedata = np.atleast_2d(linedata)
                    linedata = pd.DataFrame(
                        data=[[*linedata, 0, 0]], columns=self._base_columns
                    )

                if "atom_number" in kwargs.keys():
                    linedata["atom_number"] = kwargs["atom_number"]
                elif "atom_number" not in linedata:
                    linedata["atom_number"] = np.ones(len(linedata), dtype=float)

                if "ionization" in kwargs.keys():
                    linedata["ionization"] = kwargs["ionization"]
                elif "ionization" not in linedata and "species" in linedata:
                    linedata["ionization"] = np.array(
                        [int(s[-1]) for s in linedata["species"]], dtype=float
                    )

                if "term_upper" in kwargs.keys():
                    linedata["term_upper"] = kwargs["term_upper"]
                if "term_lower" in kwargs.keys():
                    linedata["term_lower"] = kwargs["term_lower"]
                if "reference" in kwargs.keys():
                    linedata["reference"] = kwargs["reference"]
                if "error" in kwargs.keys():
                    linedata["error"] = kwargs["error"]

        #:{"short", "long"}: Defines how much information is available
        self.lineformat = lineformat
        #:pandas.DataFrame: DataFrame that contains all the data
        self._lines = linedata  # should have all the fields (20)
        if medium in ["air", "vac", None]:
            self._medium = medium
        else:
            raise ValueError(
                f"Medium not recognized, expected one of ['air', 'vac'] , but got {medium} instead."
            )

        self.citation_info = ""
        if "citation_info" in kwargs.keys():
            self.citation_info = kwargs["citation_info"]

    def __len__(self):
        return len(self._lines)

    def __str__(self):
        return str(self._lines)

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        return self._lines.itertuples(index=False)

    def __getitem__(self, index):
        if isinstance(index, str) and hasattr(self, index):
            return getattr(self, index)
        if isinstance(index, (list, str)):
            if len(index) == 0:
                return LineList(
                    self._lines.iloc[[]],
                    lineformat=self.lineformat,
                    medium=self.medium,
                )
            values = self._lines[index].values
            if index in self.string_columns:
                values = values.astype(str)
            return values
        else:
            if isinstance(index, int):
                index = slice(index, index + 1)
            # just pass on a subsection of the linelist data, but keep it a linelist object
            return LineList(
                self._lines.iloc[index], self.lineformat, medium=self.medium
            )

    def __getattribute__(self, name):
        if name[0] != "_" and name not in dir(self) and name in self._lines:
            return self._lines[name].values
        return super().__getattribute__(name)

    @property
    def columns(self):
        return self._lines.columns

    @property
    def medium(self):
        return self._medium

    @medium.setter
    def medium(self, value):
        if self._medium is None:
            logger.warning(
                "No medium was defined for the linelist."
                " The new value of %s is assumed to be the native medium of the linelist. No conversion was performed"
            )
            self._medium = value
        if self._medium == value:
            return
        else:
            if self._medium == "air" and value == "vac":
                self._lines["wlcent"] = air2vac(self._lines["wlcent"])
                self._medium = "vac"
            elif self._medium == "vac" and value == "air":
                self._lines["wlcent"] = vac2air(self._lines["wlcent"])
                self._medium = "air"
            else:
                raise ValueError(
                    f"Type of medium not undertstood. Expected one of [vac, air], but got {value} instead"
                )

    @property
    def species(self):
        """list(str) of size (nlines,): Species name of each line"""
        return self._lines["species"].values.astype("U")

    @property
    def lulande(self):
        """list(float) of size (nlines, 2): Lower and Upper Lande factors"""
        if self.lineformat == "short":
            raise AttributeError(
                "Lower and Upper Lande Factors are only available in the long line format"
            )

            # additional data arrays for sme
        names = ["lande_lower", "lande_upper"]
        return self._lines.reindex(columns=names).values

    @property
    def extra(self):
        """list(float) of size (nlines, 3): additional line level information for NLTE calculation"""
        if self.lineformat == "short":
            raise AttributeError("Extra is only available in the long line format")
        names = ["j_lo", "e_upp", "j_up"]
        return self._lines.reindex(columns=names).values

    @property
    def atomic(self):
        """list(float) of size (nlines, 8): Data array passed to C library, should only be used for this purpose"""
        names = [
            "atom_number",
            "ionization",
            "wlcent",
            "excit",
            "gflog",
            "gamrad",
            "gamqst",
            "gamvw",
        ]
        # Select fields
        values = self._lines.reindex(columns=names).values
        values = values.astype(float)
        return values

    @property
    def index(self):
        return self._lines.index

    def sort(self, field="wlcent", ascending=True):
        """Sort the linelist

        The underlying datastructure will be replaced,
        i.e. any references will not be sorted or updated

        Parameters
        ----------
        field : str, optional
            Field to use for sorting (default: "wlcent")
        ascending : bool, optional
            Wether to sort in ascending or descending order (default: True)

        Returns
        -------
        self : LineList
            The sorted LineList object
        """

        self._lines = self._lines.sort_values(by=field, ascending=ascending)
        return self

    def add(self, species, wlcent, excit, gflog, gamrad, gamqst, gamvw):
        """Add a new line to the existing linelist

        This replaces the underlying datastructure,
        i.e. any references (atomic, etc.) will not be updated

        Parameters
        ----------
        species : str
            Name of the element and ionization
        wlcent : float
            central wavelength
        excit : float
            excitation energy in eV
        gflog : float
            gf logarithm
        gamrad : float
            broadening factor radiation
        gamqst : float
            broadening factor qst
        gamvw : float
            broadening factor van der Waals
        """

        linedata = {
            "species": species,
            "wlcent": wlcent,
            "excit": excit,
            "gflog": gflog,
            "gamrad": gamrad,
            "gamqst": gamqst,
            "gamvw": gamvw,
        }
        self._lines = self._lines.append([linedata])

    def append(self, linelist: "LineList"):
        """
        Append a linelist to this one

        Note this replaces the underlying data
        and sorts the lines by wavelength

        Parameters
        ----------
        linelist : LineList
            other linelist to append

        Returns
        -------
        self : LineList
            this object, now with appended data
        """
        if self.medium != linelist.medium:
            logger.warning(
                "The appended linelist, has its wavelength in a different medium"
            )
        if self.lineformat != linelist.lineformat:
            raise ValueError(
                (
                    "The formats of the linelists do not match. This linelist "
                    "has format %s, but the other has format %s"
                ),
                self.lineformat,
                linelist.lineformat,
            )
        self._lines = self._lines.append(linelist._lines)
        self.sort()
        return self

    def trim(self, wave_min, wave_max, rvel=None):
        """Remove lines from the linelist outside the specified
        wavelength range

        Parameters
        ----------
        wave_min : float
            lower wavelength limit in Angstrom
        wave_max : float
            upper wavelength limit in Angstrom
        rvel : float, optional
            add an additional buffer on each side, corresponding to this radial velocity, by default None

        Returns
        -------
        LineList
            trimmed linelist
        """
        if rvel is not None:
            # Speed of light in km/s
            c_light = constants.c * 1e3
            wave_min *= np.sqrt((1 - rvel / c_light) / (1 + rvel / c_light))
            wave_max *= np.sqrt((1 + rvel / c_light) / (1 - rvel / c_light))
        selection = self._lines["wlcent"] > wave_min
        selection &= self._lines["wlcent"] < wave_max
        if not np.any(selection):
            logger.warning("Trimmed linelist is empty")
        return LineList(
            self._lines[selection],
            lineformat=self.lineformat,
            medium=self.medium,
            citation_info=self.citation_info,
        )

    def cull(self, minimum_depth):
        """Remove lines from the linelist that are weaker than the cutoff

        The linedepth is an estimate and not accurate for the final stellar
        parameters, so final line depths might differ from depths in the linelist

        Parameters
        ----------
        minimum_depth : float
            the cutoff for lines to keep them. The cutoff is specified within
            the normalised spectrum, so should be between 0 and 1

        Returns
        -------
        LineList
            the culled linelist
        """
        if minimum_depth < 0 or minimum_depth > 1:
            raise ValueError(
                "minimum_depth must be between 0 and 1, but got %i",
                minimum_depth,
            )
        depth = self._lines["depth"]
        selection = depth > minimum_depth
        if not np.any(selection):
            logger.warning("Culled linelist is empty")
        return LineList(
            self._lines[selection],
            lineformat=self.lineformat,
            medium=self.medium,
            citation_info=self.citation_info,
        )

    def cull_percentage(self, percentage_to_keep):
        """Remove a percentage of the lines in the linelist, removing the
        weakest lines first

        The linedepth is an estimate and not accurate for the final stellar
        parameters, so final line depths might differ from depths in the linelist

        The cut is not exact in the number of lines, e.g. if there are many lines
        with the same depth as the cutoff, then the resulting Linelist will have
        slightly more than half the lines.

        Parameters
        ----------
        percentage_to_keep : float
            The percentage to keep in the linelist, should be between 0 and 100

        Returns
        -------
        LineList
            the culled linelist

        Raises
        ------
        ValueError
            if the percentage is not between 0 and 100
        """
        if percentage_to_keep < 0 or percentage_to_keep > 100:
            raise ValueError(
                "percentage_to_keep must be between 0 and 100, but got %i",
                percentage_to_keep,
            )
        depth = self._lines["depth"]
        cutoff = np.percentile(depth, 100 - percentage_to_keep)
        selection = depth >= cutoff
        if not np.any(selection):
            logger.warning("Culled linelist is empty")
        return LineList(
            self._lines[selection],
            lineformat=self.lineformat,
            medium=self.medium,
            citation_info=self.citation_info,
        )

    def to_dict(self):
        data = {
            "lineformat": self.lineformat,
            "medium": self.medium,
            "citation_info": self.citation_info,
        }

        for col in self._lines.columns:
            value = self._lines[col].values
            if value.dtype == object:
                value = value.astype(str)
            data[col] = value

        return data

    @classmethod
    def from_dict(cls, data):
        df_data = {
            k: v
            for k, v in data.items()
            if k not in ["lineformat", "medium", "citation_info"]
        }
        df = pd.DataFrame.from_dict(data=df_data)
        obj = cls(
            linedata=df,
            lineformat=data["lineformat"],
            medium=data["medium"],
            citation_info=data["citation_info"],
        )
        return obj

    def _save(self):
        header = {
            "lineformat": self.lineformat,
            "medium": self.medium,
            "citation_info": self.citation_info,
        }
        data = self._lines
        ext = JSONTableExtension(header, data)
        return ext

    @classmethod
    def _load(cls, ext: JSONTableExtension):
        ll = cls(ext.data, **ext.header)
        return ll

    def _save_v1(self, file, folder="linelist"):
        if folder != "" and folder[-1] != "/":
            folder = folder + "/"

        info = {
            "format": self.lineformat,
            "medium": self.medium,
            "citation_info": self.citation_info,
        }
        file.writestr(f"{folder}info.json", json.dumps(info))

        lines = self._lines.reset_index(drop=True)
        # Eventually feather should be stable, at that point we can use this
        # Until then use JSON?
        # b = io.BytesIO()
        # lines.to_feather(b)
        # file.writestr(f"{folder}data.feather", b.getvalue())
        linedata = lines.to_json(orient="records")
        file.writestr(f"{folder}data.json", linedata)

    @staticmethod
    def _load_v1(file, names, folder=""):
        for name in names:
            if name.endswith("info.json"):
                info = file.read(name)
                info = json.loads(info)
                lineformat = info["format"]
                medium = info.get("medium")
                citation_info = info.get("citation_info", "")
            elif name.endswith("data.feather"):
                b = io.BytesIO(file.read(name))
                linedata = pd.read_feather(b)
            elif name.endswith("data.json"):
                b = io.BytesIO(file.read(name))
                linedata = pd.read_json(b, orient="records")
            elif name.endswith("data.npy"):
                b = io.BytesIO(file.read(name))
                linedata = np.load(b)
                linedata = pd.DataFrame.from_records(linedata)

        return LineList(
            linedata,
            lineformat=lineformat,
            medium=medium,
            citation_info=citation_info,
        )



# -*- coding: utf-8 -*-
# TODO implement synthesis tests
import numpy as np
import pytest

from pysme.iliffe_vector import Iliffe_vector
from pysme.synthesize import synthesize_spectrum


def test_synthesis_simple(sme_2segments):
    sme = sme_2segments
    sme2 = synthesize_spectrum(sme)

    # Check if a result is there it has the expected data type
    assert sme2.synth is not None
    assert isinstance(sme2.synth, Iliffe_vector)
    assert np.all(sme2.synth.ravel() != 0)

    assert sme.wave is not None
    assert isinstance(sme2.wave, Iliffe_vector)
    assert np.issubdtype(sme2.wave.dtype, np.floating)
    assert np.all(sme2.wave.ravel() != 0)

    assert sme.spec is None


def test_synthesis_segment(sme_2segments):
    sme = sme_2segments
    # Out of range
    with pytest.raises(IndexError):
        synthesize_spectrum(sme, segments=[3])

    with pytest.raises(IndexError):
        synthesize_spectrum(sme, segments=[-1])

    sme2 = synthesize_spectrum(sme, segments=[0])
    assert len(sme2.synth[0]) != 0
    assert len(sme2.synth[1]) == 0

    assert len(sme2.wave[0]) != 0
    assert len(sme2.wave[1]) == 0

    assert sme2.wave.shape[0] == 2
    assert sme2.wave.shape[1][1] == 0

    orig = np.copy(sme2.synth[0])
    sme2 = synthesize_spectrum(sme2, segments=[1])

    assert sme2.wave.shape[1][0] != 0
    assert sme2.wave.shape[1][1] != 0

    assert np.all(sme2.synth[0] == orig)



# -*- coding: utf-8 -*-
# TODO implement NLTE tests

import os
import tempfile
from os.path import dirname

import numpy as np
import pytest

from pysme.abund import Abund
from pysme.linelist.vald import ValdFile
from pysme.nlte import DirectAccessFile, nlte
from pysme.sme import SME_Structure as SME_Struct
from pysme.sme_synth import SME_DLL
from pysme.synthesize import Synthesizer, synthesize_spectrum

from .test_largefilestorage import lfs_atmo, lfs_nlte, skipif_lfs

cwd = dirname(__file__)


def make_minimum_structure():
    sme = SME_Struct()
    sme.teff = 5000
    sme.logg = 4.4
    sme.vmic = 1
    sme.vmac = 1
    sme.vsini = 1
    sme.abund = Abund.solar()
    sme.linelist = ValdFile("{}/testcase3.lin".format(cwd))
    sme.atmo.source = "marcs2012p_t2.0.sav"
    sme.atmo.method = "grid"

    sme.wran = [[6436, 6444]]

    return sme


def test_activate_nlte():
    sme = make_minimum_structure()

    # Make sure nothing is set yet
    assert len(sme.nlte.elements) == 0

    # Add an element
    sme.nlte.set_nlte("Ca", "marcs2012p_t1.0_Ca.grd")
    assert len(sme.nlte.elements) == 1
    assert "Ca" in sme.nlte.elements

    # Add it again, shouldn't change anything
    sme.nlte.set_nlte("Ca", "marcs2012p_t1.0_Ca.grd")
    assert len(sme.nlte.elements) == 1
    assert "Ca" in sme.nlte.elements

    # Try to remove something else
    sme.nlte.remove_nlte("Na")
    assert len(sme.nlte.elements) == 1
    assert "Ca" in sme.nlte.elements

    # Remove the original element
    sme.nlte.remove_nlte("Ca")
    assert len(sme.nlte.elements) == 0

    # Add a element with a custom grid
    sme.nlte.set_nlte("Na", "test_grid.grd")
    assert len(sme.nlte.elements) == 1
    assert "Na" in sme.nlte.elements
    assert sme.nlte.grids["Na"] == "test_grid.grd"

    # Update custom grid
    sme.nlte.set_nlte("Na", "test_grid2.grd")
    assert len(sme.nlte.elements) == 1
    assert sme.nlte.grids["Na"] == "test_grid2.grd"

    # Add element without default grid
    with pytest.raises(ValueError):
        sme.nlte.set_nlte("U")

    # with a grid it should work
    sme.nlte.set_nlte("U", "test_grid.grd")
    assert sme.nlte.grids["U"] == "test_grid.grd"


@skipif_lfs
def test_run_with_nlte():
    # NOTE sme structure must have long format for NLTE
    sme = make_minimum_structure()
    sme.nlte.set_nlte("Ca", "marcs2012p_t1.0_Ca.grd")

    sme2 = synthesize_spectrum(sme)

    assert isinstance(sme2.nlte.flags, np.ndarray)
    assert np.issubdtype(sme2.nlte.flags.dtype, np.dtype("bool"))
    assert len(sme2.nlte.flags) == len(sme2.linelist)
    assert np.any(sme2.nlte.flags)


@skipif_lfs
@pytest.mark.usefixtures("lfs_atmo", "lfs_nlte")
def test_dll(lfs_atmo, lfs_nlte):
    sme = make_minimum_structure()
    elem = "Ca"
    sme.nlte.set_nlte(elem, "marcs2012p_t1.0_Ca.grd")

    libsme = SME_DLL()
    libsme.ResetDepartureCoefficients()

    syn = Synthesizer(None, lfs_atmo=lfs_atmo, lfs_nlte=lfs_nlte)
    sme = syn.get_atmosphere(sme)
    libsme.InputLineList(sme.linelist)
    libsme.InputModel(sme.teff, sme.logg, sme.vmic, sme.atmo)

    # This is essentially what update_depcoefs does, just for one element
    counter = 0
    bmat, linerefs, lineindices = nlte(sme, libsme, elem, lfs_nlte)
    for lr, li in zip(linerefs, lineindices):
        if lr[0] != -1 and lr[1] != -1:
            counter += 1
            libsme.InputDepartureCoefficients(bmat[:, lr], li)

    flags = libsme.GetNLTEflags()
    assert np.any(flags)
    assert np.count_nonzero(flags) == counter
    assert len(flags) == len(sme.linelist)

    idx = np.where(flags)[0][0]
    coeffs = libsme.GetDepartureCoefficients(idx)
    assert coeffs is not None

    # If we reset NLTE no flags should be set
    libsme.ResetDepartureCoefficients()
    flags = libsme.GetNLTEflags()
    assert not np.any(flags)
    assert len(flags) == len(sme.linelist)

    with pytest.raises(TypeError):
        libsme.InputDepartureCoefficients(None, 0)

    with pytest.raises(TypeError):
        libsme.InputDepartureCoefficients(bmat[:, [0, 1]], 0.1)

    with pytest.raises(ValueError):
        libsme.InputDepartureCoefficients([0, 1], 10)

    with pytest.raises(RuntimeError):
        libsme.InputDepartureCoefficients(bmat[:, [0, 1]], -10)


@pytest.fixture
def temp():
    file = tempfile.NamedTemporaryFile(delete=False)
    yield file.name
    try:
        os.remove(file)
    except:
        pass


def test_read_write_direct_access_file(temp: str):
    content = {
        "hello": "world",
        "I": ["have", "the", "high", "ground"],
        "teff": np.arange(100),
    }

    DirectAccessFile.write(temp, **content)
    daf = DirectAccessFile(temp)

    for key, value in content.items():
        vf = daf[key]
        if np.issubdtype(vf.dtype, np.dtype("S")):
            vf = np.char.decode(vf)

        assert np.all(vf == value)
